#include "AutoGame.h"

bool AutoGame::LoadAI(HMODULE &handle, std::string &fliename, FuncTable &funcTable){
	handle = LoadLibraryA(fliename.c_str());

	if(handle == 0){
		int errorCode = GetLastError();
		std::cout <<  "LoadLibrary(" << fliename.c_str() << ") error:" << errorCode << std::endl;
		//printf("LoadLibrary(%s) error:%d\n",ai_file,errorCode);
		return false;
	}

	funcTable.IrenseiInitialize	= (int (*)(const IRENSEI_RULE_DESCRIPTION*,int,int))GetProcAddress(handle,"IrenseiInitialize");
	funcTable.IrenseiFinalize	= (int (*)(int,int))GetProcAddress(handle,"IrenseiFinalize");
	funcTable.IrenseiThink		= (int (*)(int,IRENSEI_BOARD_POSITION*,const IRENSEI_BOARD_POSITION*,int))GetProcAddress(handle,"IrenseiThink");

	if(funcTable.IrenseiInitialize == 0 || funcTable.IrenseiFinalize == 0 || funcTable.IrenseiThink == 0){
		int errorCode = GetLastError();
		std::cout <<  "GetProcAddress(" << fliename.c_str() << ") error:" << errorCode << std::endl;
		//printf("GetProcAddress(%s):%d\n",ai_file,errorCode);
		return false;
	}

	return true;
}

void AutoGame::PlayOneGame(int param1, int param2){
	int ret;
	IRENSEI_BOARD_POSITION result,lastMove;
	mFuncTable[0].IrenseiThink(param1,&result,0,0);

	while(1){
		//̎ł
		ret = NewMove(result.x, result.y);
		if(ret >= EMResultWin){
			if(ret >= EMResultNonSpace){
				mWhiteWins++;
			}else{
				mBlackWins++;
			}
			return;
		}
		lastMove = result;
		//̎vl
		mFuncTable[1].IrenseiThink(param2,&result,&lastMove,0);
		ret = NewMove(result.x, result.y);
		if(ret >= EMResultWin){
			if(ret >= EMResultNonSpace){
				mBlackWins++;
			}else{
				mWhiteWins++;
			}
			return;
		}
		lastMove = result;
		//̎vl
		mFuncTable[0].IrenseiThink(param1,&result,&lastMove,0);
	}
}

void AutoGame::OutputKifu(int n){
	std::stringstream ss;
	ss << mDirPath << n << ".sgf";
	std::ofstream ofs(ss.str().c_str());
	if(ofs.fail()) return;

	ofs << "(;SZ[19]";
	const char bwPrefix[2] = {'B','W'};
	for(int i = 0; i < CountMove(); i++){
		Move m = GetMove(i);
		if(i%8 == 0) ofs << std::endl;
		ofs << ";" << bwPrefix[i&1] << "[" << static_cast<char>('a'+m.x) << static_cast<char>('a'+m.y) << "]";
	}
	ofs << ")";
	ofs.close();

	//ss.str("");
	//ss.clear(stringstream::goodbit);
}

void AutoGame::PlayGame(){
	mBlackWins = mWhiteWins = 0;
	SetRule(rule);

	for(int i = 0; i < mPlayTimes; i++){
		Reset();
		int param1 = mFuncTable[0].IrenseiInitialize(&rule,EBWBlack,0);
		int param2 = mFuncTable[1].IrenseiInitialize(&rule,EBWWhite,0);

		PlayOneGame(param1, param2);

		std::cout << i+1 << "ǖ" << std::endl
			<< "(" << mBlackName << ")̏:\t" << mBlackWins << std::endl
			<< "(" << mWhiteName << ")̏:\t" << mWhiteWins << std::endl << std::endl;

		mFuncTable[0].IrenseiFinalize(param1,0);
		mFuncTable[1].IrenseiFinalize(param2,0);

		if(mOutputKifu){
			OutputKifu(i);
		}
	}

	std::ofstream ofs("result.txt");
	if(ofs.fail()) return;

	ofs << "(" << mBlackName << ")̏:\t" << mBlackWins << std::endl
		<< "(" << mWhiteName << ")̏:\t" << mWhiteWins << std::endl;

	ofs.close();
}
