#include "game.h"
#include "Board.h"
#include "StringData.h"
#include <algorithm>

#include <assert.h>

#define CAPABLE_RULE_VERSION 100	/*1.00 ȉ̂ݑΉ*/


#ifdef _DEBUG
# define ASSERT(x)	assert(x)
#else
# define ASSERT(x)	
#endif


namespace Irensei {

//												//
//				[eBeB					//


//			A			//
static int straightLength( 
	const Board * board, 
	int x, int y, 			// 肷X^[gn_
	int dx, int dy,			// 肷
	int outer,				// AɊ܂߂ȂH̐
	Game::EBlackOrWhite order
)
{
	int len = 0;
	const Board::EStone me = Game::OrderToStone( order );
	for ( ; board->Get( x, y ) == me; x+= dx, y += dy ) {
		if ( x < outer || x >= Board::MaxX - outer )
			break;
		if ( y < outer || y >= Board::MaxY - outer )
			break;
		len++;
	}
	return len;
}
//				A				//
static int longestStraight( const Board * board, Game::EBlackOrWhite order, int outer )
// ꍇ킯邱ƂŁA񐔂点BVAɒBȂƂ납͕̔KvȂ
{
	int longest = 0;
	for ( int x = outer; Board::MaxX-outer > x; x++ ) {
		for ( int y = outer; Board::MaxY-outer > y; y++ ) {
			longest = std::max( longest, straightLength( board, x, y, 1, 0, outer, order ) );	// E
			longest = std::max( longest, straightLength( board, x, y, 1, 1, outer, order ) );	// E
			longest = std::max( longest, straightLength( board, x, y, 0, 1, outer, order ) );	// 
			longest = std::max( longest, straightLength( board, x, y, 1,-1, outer, order ) );	// 
		}
	}
	return longest;
}
// x, y ŗ^̂́Aorder ̐΂̂ꏊŖ΂ȂȂ
static int longestStraight( const Board * board, int x, int y, Game::EBlackOrWhite order, int outer )
// ֒TƂ͎n_܂݁AE֌Ƃ͊܂܂Ȃ
{
	int longest = 0;
	longest = std::max( longest, 		// 
			straightLength( board, x,   y, -1, 0, outer, order ) +		// 
			straightLength( board, x+1, y, +1, 0, outer, order )		// E
	);
	longest = std::max( longest, 		// 
			straightLength( board, x, y,   0, -1, outer, order ) +		// 
			straightLength( board, x, y+1, 0, +1, outer, order )		// 
	);
	longest = std::max( longest, 		// E
			straightLength( board, x,   y,   -1, -1, outer, order ) +	// 
			straightLength( board, x+1, y+1, +1, +1, outer, order )		// E
	);
	longest = std::max( longest, 		// Eオ
			straightLength( board, x,   y,   -1, +1, outer, order ) +	// 
			straightLength( board, x+1, y-1, +1, -1, outer, order )		// E
	);
	return longest;
}


//																		//
//								class Game								//

Game::Game() : rule( *MakeDefaultIrenseiRuleDescriptionVer100(static_cast<void*>(&IRENSEI_RULE_DESCRIPTION())) )
{
	board = new Board;
	Reset();
}

Game::Game( const Game & other )
		:
	rule( other.rule ),
	nextOrder( other.nextOrder ),
	board( other.board ),
	cntNuki( other.cntNuki ),
	moves( other.moves )
{
}


bool Game::SetRule( const IRENSEI_RULE_DESCRIPTION & newRule )
// @@ const j܂
{
	if ( ! CheckIrenseiRuleDescriptionValidity( &newRule ) )
		return false;	// Ȃ[͎󂯕tȂ
	if ( newRule.version > CAPABLE_RULE_VERSION )
		return false;	// ȐV[ɂ͑ΉłȂ

	memcpy( const_cast<void*>(reinterpret_cast<const void*>(&rule)), &newRule, sizeof(rule) );
	return true;
}

// ՖʂNAA̗p
void Game::Reset()
{
	board->Clear();
	nextOrder = EBWBlack;
	moves.clear();
	cntNuki = 0;
}



Game::EBlackOrWhite Game::NextOrder( Game::EBlackOrWhite order )
{
	if ( order == EBWNone )
		return EBWNone;
	return (order == EBWBlack) ? EBWWhite : EBWBlack;
}



Game::EMoveResult Game::NewMove( int x, int y )
{
	EMoveResult legality = CheckLegalMove( board, x, y, nextOrder );
	if ( legality > EMResultWin )
		return legality;

	board->ChangeStone( x, y, OrderToStone(nextOrder) );
	cntNuki = killStones( board, NextOrder( NextOrder() ), x, y );	// ̐΂グ
	moves.push_back( Context( Move( nextOrder, x, y, cntNuki ) ) );

	// 
	if ( legality == EMResultWin )
		return EMResultWin;

	nextOrder = NextOrder( nextOrder );
	return EMResultOK;

}

Game::Move Game::GetMove( unsigned int idx ) const
{
	if ( CountMove() <= idx )
		return Move( EBWNone, 0, 0, 0 );
	return moves[idx].move;
}

void Game::Undo( void )
{
	if ( moves.size() <= 0 )
		return;

	moves.pop_back();				// Ō̗j

	if ( moves.size() == 0 ) {		// ĵ肾AJnɖ߂
		Reset();
		return;
	}

	// 𕜌
	const Context & lastctx = moves.back();
	cntNuki = lastctx.move.cntNuki;

	nextOrder = NextOrder( lastctx.move.blackOrWhite );
}

static void make_string( StringData & strings, int pos2stringTable[ Board::MaxX ][ Board::MaxY ], Game::EBlackOrWhite order, int x, int y, const Board * board )
// clear() JԂȂ番
{
	StringData::String string;
	string.idx = strings.strings.size();
	string.color = Game::OrderToStone( order );
	StringData::MakeString( string, pos2stringTable, board, x, y );
	if ( ! string.stones.empty() )
		strings.strings.push_back( string );
}


int Game::killStones( Board* board, EBlackOrWhite order, int x, int y )		// グ΂̐Ԃ
{
	int cntDead = 0;

	StringData strings;
	if ( x == -1 ) {
		strings.Remake( board );
	} 
	else {
		int pos2stringTable[ Board::MaxX ][ Board::MaxY ];
		memset( pos2stringTable, -1, sizeof(pos2stringTable) );
		make_string( strings, pos2stringTable, order, x-1, y, board );
		make_string( strings, pos2stringTable, order, x+1, y, board );
		make_string( strings, pos2stringTable, order, x, y-1, board );
		make_string( strings, pos2stringTable, order, x, y+1, board );
	}

	// łAグ
	for ( std::vector<StringData::String>::const_iterator string = strings.strings.begin();
			string != strings.strings.end(); string++ 
	) {
		// ċz_O̘AȂ珈
		if ( string->breathPoints.empty() ) {
			// A̐F order Ȃ珈
			StringData::Position any/*StoneInString*/ = *string->stones.begin();
			if ( board->Get( any.x, any.y ) != OrderToStone(order) )
				continue;
			// グ
			std::set< StringData::Position >::const_iterator stone;
			for ( stone = string->stones.begin(); stone != string->stones.end(); stone++ ) {
				board->ChangeStone( stone->x, stone->y, Board::EStoneSpace );	// 󔒂
			}
			cntDead += string->stones.size();	// グ
		}
	}

	return cntDead;
}


Game::EMoveResult Game::CheckLegalMove( const Board * board, int x, int y, EBlackOrWhite order ) const 	// @Ȃ EMResultOK
{
	if ( board->Get(x,y) != Board::EStoneSpace )
		return EMResultNonSpace;

	// \Rɑł
	Board tmpboard( *board );
	tmpboard.ChangeStone( x, y, OrderToStone(order) );

	// A
	if ( order == EBWBlack ) {
		int length = longestStraight( &tmpboard, x, y, EBWBlack, 0/*ŊO܂Œׂ*/ );
		if ( length > rule.lineLength )
			return EMResultTooLong;
	}

	// Aq蔽

	// ȂAEł\Ȃ
	{
		int length = longestStraight( &tmpboard, x, y, order, rule.outerInvalidLines/*L͈͂܂Œׂ*/ );
		if ( order == EBWBlack ) {
			// ԂnAWXg
			if ( length == rule.lineLength )
				return EMResultWin;
		}
		else {
			// ԂnAȏ
			if ( length >= rule.lineLength )
				return EMResultWin;
		}
	}

	int cntKilled;

	// Ei\RɑłĂ܂j
	{
		// Ƃ͑łĂ̂ŁAɎグĂ݂
		cntKilled = killStones( &tmpboard, NextOrder( order ), x, y );	// ̐΂グ

		// ̐΂グ悤ƂĂ݂î锻łE肩ǂj
		{
				StringData strings;
				int pos2stringTable[ Board::MaxX ][ Board::MaxY ];
				memset( pos2stringTable, -1, sizeof(pos2stringTable) );
				make_string( strings, pos2stringTable, order, x, y, &tmpboard );
				ASSERT( ! strings.strings.empty() );
				if ( strings.strings.front().breathPoints.empty() )
					return EMResultSuicide;		// ꂽI
		}
	}

	// i|kLŁȂԂ|kLŁA肪ł΂ʂȂ_j
	//iQȏƂȂA̕܂ނƂ͖j
	if ( cntKilled == 1 && cntNuki == 1 ) {
		Move last = moves.back().move;	// ponNuki  true ȂAempty() ͕KvȂ
		if ( tmpboard.Get( last.x, last.y ) == Board::EStoneSpace )
			return EMResultKou;		// ł
	}

	// łĂ܂
	return EMResultOK;
}







}	// Irensei

